# -*- coding: utf-8 -*-
#
#   Copyright (C) 2018 European Synchrotron Radiation Facility, Grenoble, France
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


__authors__ = ["W. De Nolf"]
__license__ = "MIT"

import logging
import errno
from contextlib import contextmanager
from time import sleep
import datetime
import sys
import os

from . import oarshell
from . import oarresource
from . import timeutils

logger = logging.getLogger()


class Job(object):
    """Manage existing jobs but does not create new jobs
    """

    _POSSIBLE_STATES = ('Launching', 'Resuming', 'Finishing', 'Terminated',
                        'Error', 'Running', 'Waiting', 'Hold', 'Suspended')
    """list all the possible state the job can take"""

    def __init__(self, jobid):
        """

        :param int jobid:
        """
        self._jobid = jobid
        self._fixed_stats = None
        """Used to store stats when using the `fixed_stats` context
        manager"""

    @property
    def jobid(self):
        return self._jobid

    @property
    def exists(self):
        return bool(self.status)

    @property
    def is_terminated(self):
        return self.status == 'Terminated'

    @property
    def has_error(self):
        return self.status == 'Error'

    @property
    def is_finished(self):
        return self.status in ['Terminated', 'Error']

    @property
    def is_running(self):
        return self.status == 'Running'

    @property
    def is_waiting(self):
        return self.status == 'Waiting'

    @property
    def is_intermediate(self):
        return self.status in ['Launching', 'Resuming', 'Finishing']

    @property
    def needsresume(self):
        return self.status in ['Hold', 'Suspended']

    @property
    def status(self):
        if self._fixed_stats:
            return self['state']
        else:
            if self.jobid is None:
                return None
            else:
                return oarshell.jobstatus(self.jobid)

    @property
    def stats(self):
        if self._fixed_stats:
            return self._fixed_stats
        else:
            if self.jobid is None:
                stats = {}
            else:
                stats, err, code = oarshell.jobstats(self.jobid)
                self._raise_if_error(err, code, 'Cannot retrieve job stats')
                self._parse_stats(stats)
            return stats

    @contextmanager
    def fixed_stats(self):
        first = not bool(self._fixed_stats)
        if first:
            self._fixed_stats = self.stats
        yield
        if first:
            self._fixed_stats = None

    def __getitem__(self, key):
        return self.stats.get(key, None)

    def wait(self, states=('Terminated', 'Error'), timeout=None, refresh=1,
             silent=False):
        """

        :param states: state or states we are waiting for
        :type states: str or tuple
        :param timeout: if not none, second before the timeout.
        :type timeout: int or None
        :param int refresh: time (in seconds) between two observations of the job
                            state
        :param silent: if False then write to stdout advancement ('.')
        """
        # make sure _until will always be iterable (deal with str, bytes...)
        _states = states
        if isinstance(_states, (tuple, list, dict)) is False:
            _states = (_states,)
        for _state in _states:
            if _state not in self._POSSIBLE_STATES:
                raise ValueError('%s is not a valid state. Unable to wait for it' % _state)

        _until = lambda: self.status in _states

        waited_time = 0
        newline = False
        while not _until():
            sleep(refresh)
            waited_time = waited_time + refresh
            if timeout is not None:
                if waited_time >= timeout:
                    logger.warning('wait on job %s has reach a timeout' % self.jobid)
                    return
            if not silent:
                sys.stdout.write('.')
                sys.stdout.flush()
                newline = True
        if newline:
            print('')

    @property
    def runtime(self):
        """Effective execution time, excluding queue time
        """
        dt = self._time_diff('stopTime', 'startTime')
        return max(dt, datetime.timedelta(seconds=0))

    @property
    def time_scheduled(self):
        """Time this job was scheduled, excluding queue and runtime
        """
        dt = self._time_diff('startTime', 'scheduledStart')
        return max(dt, datetime.timedelta(seconds=0))

    @property
    def time_enqueued(self):
        """Time this job was enqueue (not scheduled for execution)
        """
        dt = self._time_diff('scheduledStart', 'submissionTime')
        return max(dt, datetime.timedelta(seconds=0))

    @property
    def time_to_start(self):
        """Time until the job starts
        """
        dt = self._time_diff('startTime', timeutils.now())
        return max(dt, datetime.timedelta(seconds=0))

    @property
    def stderr_file(self):
        return self._std_file('stderr_file')

    @property
    def stdout_file(self):
        return self._std_file('stdout_file')

    @property
    def stderr(self):
        return self._std_read(self.stderr_file)

    @property
    def stdout(self):
        return self._std_read(self.stdout_file)

    def _std_file(self, filename):
        with self.fixed_stats():
            filename = self[filename]
            if os.path.dirname(filename):
                return filename
            else:
                return os.path.join(self.working_directory, filename)

    @staticmethod
    def _std_read(filename):
        try:
            with open(filename, mode='r') as f:
                return f.read()
        except IOError as e:
            if e.errno == errno.ENOENT:
                return None
            else:
                raise e

    def remove_logs(self):
        with self.fixed_stats():
            for f in [self.stderr_file, self.stdout_file]:
                try:
                    os.remove(f)
                except OSError as e:
                    if e.errno == errno.ENOENT:
                        pass
                    else:
                        raise e

    @property
    def exit_code(self):
        """

        :return: * None: interrupted by user
                 * ==0: success
                 * !=0: error
        """
        return self['exit_code']

    @property
    def name(self):
        return self['name']

    @property
    def project(self):
        return self['project']

    @property
    def command(self):
        return self['command']

    @property
    def working_directory(self):
        return self['launchingDirectory']

    @property
    def log_directory(self):
        with self.fixed_stats():
            path = os.path.dirname(self.stdout_file)
            if path:
                return path
            else:
                return self.working_directory

    @property
    def log_base(self):
        with self.fixed_stats():
            base = os.path.basename(self.stdout_file)
            base = base.split('.')
            if base[-1] == 'stdout':
                base.pop()
            name = self.name
            base = ['%jobname%' if s == name else s
                    for s in base if s]
            name = str(self.jobid)
            base = ['%jobid%' if s == name else s
                    for s in base]
            return '.'.join(base)

    @property
    def resource(self):
        with self.fixed_stats():
            cmd = self['initial_request']
            if cmd:
                properties = None
            else:
                cmd = self['wanted_resources']
                properties = self['properties']
            return oarresource.Resource.from_cli(cmd, properties)

    def __str__(self):
        try:
            with self.fixed_stats():
                stats = self.stats
                lst = []
                for key in ['name', 'project', 'state', 'owner']:
                    lst.append('{} = {}'.format(key, stats[key]))
                lst.append('runtime = {}'.format(self.runtime))
                return 'Job({})\n '.format(self.jobid) + '\n '.join(lst)
        except RuntimeError:
            return 'Job(non existing {})'.format(self.jobid)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.jobid)

    def resume(self):
        """Resume suspended job or schedule job for execution
        """
        out, err, code = oarshell.jobresume(self.jobid)
        self._raise_if_error(err, code, 'Cannot resume job')

    def suspend(self):
        """Remove job from schedule or suspend running job
        (currently not permitted)
        """
        out, err, code = oarshell.jobhold(self.jobid)
        self._raise_if_error(err, code, 'Cannot suspend job')

    def interrupt(self):
        """Stop a running, waiting or holding job
        """
        out, err, code = oarshell.jobdel(self.jobid)
        self._raise_if_error(err, code, 'Cannot suspend job')

    stop = interrupt

    @property
    def definition(self):
        with self.fixed_stats():
            kwargs = {}
            kwargs['command'] = self.command
            kwargs['name'] = self.name
            kwargs['project'] = self.project
            kwargs['working_directory'] = self.working_directory
            if self.log_directory != self.working_directory:
                kwargs['log_directory'] = self.log_directory
            kwargs['log_base'] = self.log_base
            kwargs['resource'] = self.resource
            return JobFactory(**kwargs)

    @staticmethod
    def _parse_stats(stats):
        # Dates are in local timezone
        for timekey in ['startTime', 'stopTime', 'scheduledStart', 'submissionTime']:
            value = stats.get(timekey, 0)
            if value:
                stats[timekey] = timeutils.fromtimestamp(value)
            else:
                stats[timekey] = value

        value = stats.get('walltime', None)
        if value:
            stats['walltime'] = datetime.timedelta(seconds=int(value))

    def _time_diff(self, end, start):
        """
        
        :param end: None (now)
        :type end: str or datetime or None
        :param start: None (now)
        :type start: str or datetime or None
        :return timedelta: 
        """
        with self.fixed_stats():
            return self._get_time(end) - self._get_time(start)

    def _get_time(self, tm):
        """
        
        :param tm: None (now)
        :type tm: str or datetime or None
        :return datetime: 
        """
        if isinstance(tm, datetime.datetime):
            return tm
        else:
            tm = self[tm]
            if not tm:
                tm = timeutils.now()
            return tm

    def _raise_if_error(self, err, code, msg):
        if code:
            errname = errno.errorcode.get(code, None)
            msg = '{} (Jobid={},Error={},{})\n{}'.format(msg, self.jobid, code, errname, err)
            logger.error(msg)
            raise RuntimeError(msg)

    def __eq__(self, other):
        return self.jobid == other.jobid

    def __ne__(self, other):
        return self.jobid != other.jobid

    def __lt__(self, other):
        return self.jobid < other.jobid

    def __gt__(self, other):
        return self.jobid > other.jobid

    def __le__(self, other):
        return self.jobid <= other.jobid

    def __ge__(self, other):
        return self.jobid >= other.jobid


class JobFactory(object):
    """Define and submit new jobs
    """

    def __init__(self, command=None, resource=None, name=None, project=None,
                 working_directory=None, log_directory=None, log_base=None, **resource_parameters):
        self._name = None
        self._log_base = None
        self.command = command
        if resource_parameters:
            if resource:
                logger.warning('Skip resource parameters {}'.format(resource_parameters))
            else:
                resource = oarresource.Resource(**resource_parameters)
        self.resource = resource
        self.name = name
        self.project = project
        self.working_directory = working_directory
        self.log_directory = log_directory
        self.log_base = log_base

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self._default_name()

    @property
    def log_base(self):
        return self._log_base

    @log_base.setter
    def log_base(self, value):
        if value:
            self._log_base = value
        else:
            self._log_base = '%jobname%.%jobid%'
        self._default_name()

    def _default_name(self):
        if not self.name and self.log_base:
            if '%jobname%' in self.log_base:
                self.name = 'OAR'

    @property
    def cli_arguments(self):
        args = []
        if self.name:
            args += ['-n', self.name]
        if self.project:
            args += ['--project', self.project]
        if self.working_directory:
            path = os.path.expanduser(self.working_directory)
            args += ['-d', path]
        log_base = self.log_base
        if self.log_directory:
            path = os.path.expanduser(self.log_directory)
            log_base = os.path.join(path, log_base)
        args += ['-O', log_base + '.stdout']
        args += ['-E', log_base + '.stderr']
        if self.resource:
            add = self.resource.cli_arguments
            if add:
                args += add
        if self.command:
            args.append(self.command)
        return args

    @property
    def cli_string(self):
        return oarshell.cli_args2str(*self.cli_arguments)

    def __eq__(self, other):
        return self.cli_arguments == other.cli_arguments

    def __ne__(self, other):
        return self.cli_arguments != other.cli_arguments

    def __str__(self):
        return self.cli_string

    def __nonzero__(self):
        return self.__bool__()

    def __bool__(self):
        return bool(self.cli_arguments)

    def submit(self, hold=False):
        if not self.command:
            msg = 'Cannot submit a job without a command'
            logger.error(msg)
            raise RuntimeError(msg)

        cli_args = self.cli_arguments
        if hold:
            cli_args.insert(0, '--hold')
        jobid, err, code = oarshell.oarsub(*cli_args)
        if code:
            errname = errno.errorcode.get(code, None)
            msg = 'Cannot submit job (Error={}:{})\n{}'.format(code, errname, err)
            logger.error(msg)
            raise RuntimeError(msg)
        return Job(jobid)


def submit(hold=False, **parameters):
    jobdef = JobFactory(**parameters)
    return jobdef.submit(hold=hold)


def search(name=None, project=None, owner=None, state=None,
           start=None, end=None, **properties):
    """

    :param str name: 
    :param str project: 
    :param str owner: 
    :param datetime start: 
    :param datetime end: 
    :param str state: 
    :return list: list of jobs
    """
    # https://github.com/oar-team/oar/blob/2.5/sources/core/database/mysql_structure.sql
    sqlquery = []
    if name:
        sqlquery.append("job_name='{}'".format(name))
    if project:
        sqlquery.append("project='{}'".format(project))
    if owner:
        sqlquery.append("job_user='{}'".format(owner))
    if state:
        sqlquery.append("state='{}'".format(state))
    if start:
        start = timeutils.totimestamp(start)
        sqlquery.append("start_time>='{}'".format(start))
    if end:
        end = timeutils.totimestamp(end)
        sqlquery.append("stop_time<='{}'".format(end))
    for k, (op, v) in properties.items():
        sqlquery.append("{}{}'{}'".format(k, op, v))
    if not sqlquery:
        return []

    result, err, code = oarshell.oarstat('--sql', ' and '.join(sqlquery))
    if code:
        # json parsing gives error when query returns no results
        # (bug in OAR perl library) so don't raise an exception here
        errname = errno.errorcode.get(code, None)
        msg = 'Job search error {} ({})\n{}'.format(code, errname, err)
        logger.debug(msg)
        # raise RuntimeError(msg)
        return []
    if result:
        return [Job(jobid) for jobid in result]
    else:
        return []



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2018 European Synchrotron Radiation Facility, Grenoble, France
#
#   Principal author:   Wout De Nolf (wout.de_nolf@esrf.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import unittest
import os
import tempfile
from time import sleep
import shutil

from .. import oarjob
from ..oarresource import Resource
from ..oarshell import oarinstalled


class test_oarjob(unittest.TestCase):
    def setUp(self):
        self.working_directory = tempfile.mkdtemp(dir=os.getcwd())

    def tearDown(self):
        shutil.rmtree(self.working_directory)

    def skipoar(self):
        if not oarinstalled():
            raise unittest.SkipTest("oar cli tools not available")

    def test_status(self):
        self.skipoar()
        job = oarjob.Job(1)
        with job.fixed_stats():
            self.assertTrue(job.exists)
            self.assertTrue(job.is_finished)
            self.assertFalse(job.is_running)
            self.assertFalse(job.is_waiting)
            self.assertFalse(job.is_intermediate)
            self.assertFalse(job.needsresume)
        self.assertEqual(job['state'], job.status)

    def test_jobfactory(self):
        self.skipoar()
        self.assertTrue(oarjob.JobFactory())
        self.assertTrue(oarjob.JobFactory(command='ls -all'))

        res = Resource(nodes=1, cpu=2, core=2, gpu=True, mem_core_mb=8000)
        info = oarjob.JobFactory(resource=res, command='ls -all', log_base='test')
        cli = "-O test.stdout -E test.stderr -l nodes=1/cpu=2/core=2 -p \"gpu='YES' and mem_core_mb>=8000\" \"ls -all\""
        self.assertEqual(cli, info.cli_string)

    def definition(self, seconds, name):
        command = 'python -c "from time import sleep\nfor i in range({}):\n print(i)\n sleep(1)"'
        resource = Resource(core=1, walltime={'seconds': seconds * 3})
        jobdef = oarjob.JobFactory(name=name, project='oarpy',
                                   resource=resource, command=command.format(seconds),
                                   working_directory=self.working_directory)
        output = list(range(seconds))
        return jobdef, output

    def _check_output(self, job, expected):
        output = [int(i) for i in job.stdout.split('\n') if i]
        self.assertEqual(output, expected)

    def test_immediate(self):
        self.skipoar()
        jobdef, expected = self.definition(5, 'immediate')
        job = jobdef.submit()
        self.assertEqual(job.definition, jobdef)
        job.wait(silent=True)
        self.assertEqual(job.exit_code, 0)
        self._check_output(job, expected)

    def test_notimmediate(self):
        self.skipoar()
        jobdef, expected = self.definition(5, 'notimmediate')
        job = jobdef.submit(hold=True)
        job.wait(states=('Hold', 'Suspended'), silent=True)
        job.resume()
        job.wait(silent=True)
        self.assertEqual(job.exit_code, 0)
        self._check_output(job, expected)

    def test_interrupt(self):
        self.skipoar()
        jobdef, expected = self.definition(60, 'interrupt')
        job = jobdef.submit()
        job.wait(states=('Running', 'Terminated', 'Error'), silent=True)
        sleep(5)
        job.interrupt()
        job.wait(silent=True)
        self.assertEqual(job.exit_code, None)

    def test_suspend(self):
        self.skipoar()
        jobdef, expected = self.definition(60, 'interrupt')
        job = jobdef.submit()
        job.wait(states=('Running', 'Terminated', 'Error'), silent=True)
        self.assertRaises(RuntimeError, job.suspend)
        # job.wait(states='Hold', silent=True)
        # job.resume()
        # job.wait(silent=True)
        # self.assertEqual(job.exit_code,0)
        # self._check_output(job,expected)


def test_suite():
    """Test suite including all test suites"""
    testSuite = unittest.TestSuite()
    testSuite.addTest(test_oarjob("test_status"))
    testSuite.addTest(test_oarjob("test_jobfactory"))
    testSuite.addTest(test_oarjob("test_immediate"))
    testSuite.addTest(test_oarjob("test_notimmediate"))
    testSuite.addTest(test_oarjob("test_interrupt"))
    testSuite.addTest(test_oarjob("test_suspend"))
    return testSuite


if __name__ == '__main__':
    import sys

    mysuite = test_suite()
    runner = unittest.TextTestRunner()
    if not runner.run(mysuite).wasSuccessful():
        sys.exit(1)



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2018 European Synchrotron Radiation Facility, Grenoble, France
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__authors__ = ["W. De Nolf"]
__license__ = "MIT"

import datetime
import dateutil.tz


def astimezone(dt, tz):
    try:
        return dt.astimezone(tz)
    except ValueError:
        return dt.replace(tzinfo=tz)


tzlocal = dateutil.tz.tzlocal()
tzutc = dateutil.tz.tzutc()
epoch = astimezone(datetime.datetime(1970, 1, 1), tzutc)


def totimestamp(dt):
    """
    :param datetime dt: assume local tz when tz-unaware
    :return int: seconds since UNIX epoch
    """
    return int(round((astimezone(dt, tzlocal) - epoch).total_seconds()))


def fromtimestamp(stamp, tz=tzlocal):
    """
    :param int stamp: seconds since UNIX epoch
    :param tzinfo tz: local by default (Optional)
    :return datetime: tz-aware
    """
    return datetime.datetime.fromtimestamp(stamp, tz)


def now(tz=tzlocal):
    """
    :param tzinfo tz: local by default
    :return datetime: tz-aware
    """
    return datetime.datetime.now(tz)


def add(dt, **kwargs):
    """
    :param datetime dt:
    :return datetime:
    """
    return dt + datetime.timedelta(**kwargs)



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2018 European Synchrotron Radiation Facility, Grenoble, France
#
#   Principal author:   Wout De Nolf (wout.de_nolf@esrf.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import unittest
from .. import oarshell


class test_utils(unittest.TestCase):

    def skipoar(self):
        if not oarshell.oarinstalled():
            raise unittest.SkipTest("oar cli tools not available")

    def test_installed(self):
        oarshell.oarinstalled()

    def test_oarstat(self):
        self.skipoar()
        out, err, code = oarshell.jobstats(1)
        self.assertTrue(isinstance(out, dict))

    def test_cli(self):
        args = '-p', "gpu=YES"
        cmd1 = oarshell.cli_args2str(*args)
        cmd2 = "-p gpu=YES"
        self.assertEqual(cmd1, cmd2)
        self.assertEqual(args, oarshell.cli_str2args(cmd2))

        args = '-p', "gpu='YES'"
        cmd1 = oarshell.cli_args2str(*args)
        cmd2 = "-p gpu='YES'"
        self.assertEqual(cmd1, cmd2)
        self.assertEqual(('-p', "gpu=YES"), oarshell.cli_str2args(cmd1))

        args = '-p', "gpu='YES' and mem_core_mb>=8000"
        cmd1 = oarshell.cli_args2str(*args)
        cmd2 = '-p "gpu=\'YES\' and mem_core_mb>=8000"'
        self.assertEqual(cmd1, cmd2)
        self.assertEqual(args, oarshell.cli_str2args(cmd2))


def test_suite():
    """Test suite including all test suites"""
    testSuite = unittest.TestSuite()
    testSuite.addTest(test_utils("test_installed"))
    testSuite.addTest(test_utils("test_oarstat"))
    testSuite.addTest(test_utils("test_cli"))
    return testSuite


if __name__ == '__main__':
    import sys

    mysuite = test_suite()
    runner = unittest.TextTestRunner()
    if not runner.run(mysuite).wasSuccessful():
        sys.exit(1)



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2018 European Synchrotron Radiation Facility, Grenoble, France
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__authors__ = ["W. De Nolf"]
__license__ = "MIT"

import datetime
import re
from . import oarshell


def walltime2str(tdelta):
    h, rem = divmod(int(round(tdelta.total_seconds())), 3600)
    m, s = divmod(rem, 60)
    return '{:02d}:{:02d}:{:02d}'.format(h, m, s)


def str2walltime(s):
    h, m, s = s.split(':')
    return datetime.timedelta(hours=int(h), minutes=int(m), seconds=int(s))


property_pattern = re.compile('(\w+)([>=<]+)\'?([\w\d]+)\'?')


# Includes the default ones (space around operator):
# property_pattern = re.compile('(\w+) *([>=<]+) *\'?([\w\d]+)\'?')

class Resource(object):
    def __init__(self, host=None, nodes=None, cpu=None, core=None, gpu=False,
                 walltime=None, mem_core_mb=None, **properties):
        """
        
        Custom properties can be defined: for example cpu_vendor=('=',INTEL)"

        :param int host: number of nodes to be used (Optional)
        :param int nodes: number of nodes to be used (Optional)
        :param int cpu: number of cpu's per node (Optional)
        :param int core: number of cores per cpu (Optional)
        :param bool gpu: need a gpu or not (Optional)
        :param walltime: hours when a number, dict keys: days, seconds,
                         microseconds, milliseconds, minutes, hours, weeks
                         (Optional)
        :type walltime: timedelta or dict or num
        :param int mem_core_mb: required memory per core in MB (Optional)
        """
        if not nodes:
            nodes = host
        self.nodes = nodes
        self.cpu = cpu
        self.core = core
        self.gpu = gpu
        self.walltime = walltime
        self.mem_core_mb = mem_core_mb
        self.properties = properties

    @property
    def walltime(self):
        return self._walltime

    @walltime.setter
    def walltime(self, value):
        if isinstance(value, datetime.timedelta) or value is None:
            self._walltime = value
        else:
            try:
                self._walltime = datetime.timedelta(**value)
            except TypeError:
                self._walltime = datetime.timedelta(hours=value)

    @property
    def memory(self):
        return self.cores * self.mem_core_mb

    @property
    def cores(self):
        n = 1
        for m in [self.nodes, self.cpu, self.core]:
            if m:
                n *= m
        return n

    @property
    def _cli_resource(self):
        resources = []
        if self.nodes:
            resources.append('nodes={}'.format(self.nodes))
        if self.cpu:
            resources.append('cpu={}'.format(self.cpu))
        if self.core:
            resources.append('core={}'.format(self.core))
        cliarg = '/'.join(resources)
        if self.walltime:
            walltime = walltime2str(self.walltime)
            if cliarg:
                cliarg = '{},walltime={}'.format(cliarg, walltime)
            else:
                cliarg = 'walltime={}'.format(walltime)
        return cliarg

    @property
    def _cli_properties(self):
        properties = []
        if self.gpu:
            properties.append("gpu='YES'")
        if self.mem_core_mb:
            properties.append("mem_core_mb>={:d}".format(self.mem_core_mb))
        for k, (op, v) in self.properties.items():
            properties.append("{}{}'{}'".format(k, op, v))
        cliarg = ' and '.join(properties)
        return cliarg

    @property
    def cli_arguments(self):
        args = []
        add = self._cli_resource
        if add:
            args += ['-l', add]
        add = self._cli_properties
        if add:
            args += ['-p', add]
        return args

    @property
    def cli_string(self):
        return oarshell.cli_args2str(*self.cli_arguments)

    @classmethod
    def from_cli(cls, cmd, properties=None):
        kwargs = {}
        cli_arguments = list(oarshell.cli_str2args(cmd))
        arg = oarshell.cli_getarg(cli_arguments, '-l')
        if arg:
            for key in ['host', 'nodes', 'cpu', 'core']:
                match = re.search('{}=(\d+)'.format(key), arg)
                if match:
                    kwargs[key] = match.groups()[0]
            match = re.search('walltime=(\d+:\d+:\d+)', arg)
            if match:
                kwargs['walltime'] = str2walltime(match.groups()[0])
        arg = oarshell.cli_getarg(cli_arguments, '-p')
        if not arg and properties:
            arg = properties
        if arg:
            for match in property_pattern.finditer(arg):
                k, op, v = match.groups()
                if k == 'gpu':
                    kwargs['gpu'] = v == 'YES'
                elif k == 'mem_core_mb':
                    kwargs['mem_core_mb'] = int(v)
                else:
                    kwargs[k] = op, v
        return cls(**kwargs)

    def __eq__(self, other):
        return self.cli_arguments == other.cli_arguments

    def __ne__(self, other):
        return self.cli_arguments != other.cli_arguments

    def __str__(self):
        return self.cli_string

    def __nonzero__(self):
        return self.__bool__()

    def __bool__(self):
        return bool(self.cli_arguments)



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2018 European Synchrotron Radiation Facility, Grenoble, France
#
#   Principal author:   Wout De Nolf (wout.de_nolf@esrf.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import unittest
from ..oarresource import Resource


class test_oarresource(unittest.TestCase):
    def test_cores(self):
        self.assertFalse(Resource())
        self.assertTrue(Resource(core=1))
        self.assertEqual(Resource(nodes=1).cores, 1)
        self.assertEqual(Resource(cpu=1).cores, 1)
        self.assertEqual(Resource(core=1).cores, 1)

    def test_to_cli(self):
        res = Resource(nodes=1, cpu=2, core=2, gpu=True, mem_core_mb=8000)
        cli = '-l nodes=1/cpu=2/core=2 -p "gpu=\'YES\' and mem_core_mb>=8000"'
        self.assertEqual(cli, res.cli_string)
        self.assertEqual(res.cores, 4)

    def test_from_cli(self):
        res = Resource()
        self.assertEqual(res, Resource.from_cli(res.cli_string))
        res = Resource(core=10)
        self.assertEqual(res, Resource.from_cli(res.cli_string))
        res = Resource(nodes=1, cpu=2, core=2, gpu=True)
        self.assertEqual(res, Resource.from_cli(res.cli_string))
        res = Resource(nodes=1, cpu=2, core=2, mem_core_mb=8000)
        self.assertEqual(res, Resource.from_cli(res.cli_string))
        res = Resource(nodes=1, cpu=2, core=2, gpu=True, mem_core_mb=8000)
        self.assertEqual(res, Resource.from_cli(res.cli_string))


def test_suite():
    """Test suite including all test suites"""
    testSuite = unittest.TestSuite()
    testSuite.addTest(test_oarresource("test_cores"))
    testSuite.addTest(test_oarresource("test_to_cli"))
    testSuite.addTest(test_oarresource("test_from_cli"))
    return testSuite


if __name__ == '__main__':
    import sys

    mysuite = test_suite()
    runner = unittest.TextTestRunner()
    if not runner.run(mysuite).wasSuccessful():
        sys.exit(1)



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2018 European Synchrotron Radiation Facility, Grenoble, France
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__authors__ = ["W. De Nolf"]
__license__ = "MIT"


import os
import subprocess
import json
import logging
import shlex

logger = logging.getLogger(__name__)


def _execute(args, **kwargs):
    logger.debug('Subprocess: {}'.format(' '.join(args)))
    proc = subprocess.Popen(args, **kwargs)
    out, err = proc.communicate()
    logger.debug('stdout: {}'.format(out))
    logger.debug('stderr: {}'.format(err))
    if out:
        out = out.decode()
    if err:
        err = err.decode()
    return out, err, proc.returncode


def installed(*args):
    try:
        devnull = open(os.devnull)
        _execute(args, stdout=devnull, stderr=devnull)
    except OSError as e:
        if e.errno == os.errno.ENOENT:
            return False
    return True


def execute(*args):
    try:
        return _execute(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError as e:
        return None, None, e.errno


def oarinstalled():
    return installed('oarsub', '--version')


def executejson(cmd, *args):
    """

    :param str cmd:
    :return tuple: out(dict), err(str or None), exitcode(int)
    """
    out, err, code = execute(cmd, '-J', *args)
    if out:
        try:
            start = out.index('{')
            stop = out.index('}')
        except ValueError:
            out = {}
        else:
            out = json.loads(out[start:stop+1])
    else:
        out = {}
    return out, err, code


def oarstat(*args):
    """

    :return tuple: out(dict), err(str or None), exitcode(int)
    """
    out, err, code = execute('oarstat', '-J', *args)
    if out:
        out = json.loads(out)
    else:
        out = {}
    return out, err, code


def oarjobstat(jobid, *args):
    """oarstat for a single job

    :param int jobid:
    :return tuple: out(str or dict or None), err(str or None), exitcode(int)
    """
    jobid = str(jobid)
    out, err, code = oarstat('-j', jobid, *args)
    if out:
        out = out.get(jobid, None)
    else:
        out = None
    return out, err, code


def oarsub(*args):
    """

    :return tuple: out(str or None), err(str or None), exitcode(int)
    """
    out, err, code = execute('oarsub', '-J', *args)
    if out:
        try:
            start = out.index('{')
            stop = out.index('}')
        except ValueError:
            out = 0
        else:
            out = json.loads(out[start:stop+1]).get('job_id', 0)
    else:
        out = 0
    return out, err, code


def jobstatus(jobid):
    """
    :param int jobid:
    :return str or None: * None: means job is not registered
                         * Hold: not scheduled (needs to be resumed)
                         * Waiting: scheduled for execution
                         * Suspended: running process is suspended
                            (needs to be resumed)
                         * Launching: process is starting
                         * Running: process is running
                         * Resuming: resuming after hold or suspended
                         * Finishing: process is stopping
                         * Terminated: process finished successfully
                         * Error: process finished successfully
    """
    out, _, code = oarjobstat(jobid, '-s')
    if code:
        return None
    else:
        return out


def jobstats(jobid):
    """
    :param int jobid:
    :return tuple: out(str or None), err(str or None), exitcode(int)
    """
    out, err, code = oarjobstat(jobid, '-f')
    if not isinstance(out, dict):
        out = {}
    return out, err, code


def jobresume(jobid):
    return execute('oarresume', str(jobid))


def jobhold(jobid):
    # Not permissions to hold running job with the -r option
    return execute('oarhold', str(jobid))


def jobdel(jobid, signal=None):
    args = []
    if signal:
        args = ['-s', signal]
    return execute('oardel', str(jobid), *args)


def cli_args2str(*cli_arguments):
    return subprocess.list2cmdline(cli_arguments)


def cli_str2args(cmd):
    return tuple(shlex.split(cmd))


def cli_getarg(cli_arguments, flag):
    if flag in cli_arguments:
        i = cli_arguments.index(flag)
        ret = cli_arguments[i+1]
        if ret.startswith('-'):
            return ''
        else:
            return ret
    else:
        return ''



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2018 European Synchrotron Radiation Facility, Grenoble, France
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from setuptools import find_packages
from sphinx.setup_command import BuildDoc
import os
import io
import sys
import logging

logger = logging.getLogger(__file__)
try:
    import sphinx
    import sphinx.util.console

    sphinx.util.console.color_terminal = lambda: False
    from sphinx.setup_command import BuildDoc
except ImportError:
    sphinx = None
try:
    from setuptools import Command

    logger.info("Use setuptools")
except ImportError:
    try:
        from numpy.distutils.core import Command
    except ImportError:
        from distutils.core import Command

cmdclass = {'build_sphinx': BuildDoc}

PROJECT = 'oarpy'

VERSION = '0.1.0b1'

classifiers = ["Development Status :: 4 - Beta",
               "Environment :: Console",
               "Intended Audience :: Education",
               "Intended Audience :: Science/Research",
               "License :: OSI Approved :: MIT License",
               "Natural Language :: English",
               "Operating System :: POSIX",
               "Programming Language :: Python :: 2.7",
               "Programming Language :: Python :: 3.4",
               "Programming Language :: Python :: 3.5",
               "Programming Language :: Python :: 3.6",
               "Topic :: Software Development :: Libraries :: Python Modules",
               ]


def get_readme():
    """Returns content of README.rst file"""
    dirname = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(dirname, "README.rst")
    with io.open(filename, "r", encoding="utf-8") as fp:
        long_description = fp.read()
    return long_description


if sphinx is not None:
    class BuildDocCommand(BuildDoc):
        """Command to build documentation using sphinx.

        Project should have already be built.
        """

        def run(self):
            # make sure the python path is pointing to the newly built
            # code so that the documentation is built on this and not a
            # previously installed version

            build = self.get_finalized_command('build')
            sys.path.insert(0, os.path.abspath(build.build_lib))

            # Build the Users Guide in HTML and TeX format
            for builder in ['html', 'latex']:
                self.builder = builder
                self.builder_target_dir = os.path.join(self.build_dir, builder)
                self.mkpath(self.builder_target_dir)
                BuildDoc.run(self)
            sys.path.pop(0)
else:
    class SphinxExpectedCommand(Command):
        """Command to inform that sphinx is missing"""
        user_options = []

        def initialize_options(self):
            pass

        def finalize_options(self):
            pass

        def run(self):
            raise RuntimeError(
                'Sphinx is required to build or test the documentation.\n'
                'Please install Sphinx (http://www.sphinx-doc.org).')


    BuildDocCommand = SphinxExpectedCommand

PACKAGES = find_packages()


def configuration(parent_package='', top_path=None):
    """Recursive construction of package info to be used in setup().

    See http://docs.scipy.org/doc/numpy/reference/distutils.html#numpy.distutils.misc_util.Configuration
    """
    try:
        from numpy.distutils.misc_util import Configuration
    except ImportError:
        raise ImportError(
            "To install this package, you must install numpy first\n"
            "(See https://pypi.python.org/pypi/numpy)")
    config = Configuration(None, parent_package, top_path)
    config.set_options(
        ignore_setup_xxx_py=True,
        assume_default_configuration=True,
        delegate_options_to_subpackages=True,
        quiet=True)
    config.add_subpackage(PROJECT)
    return config


def get_project_configuration(dry_run):
    """Returns project arguments for setup"""
    install_requires = [
    ]

    setup_requires = ["setuptools"]

    package_data = {

    }
    entry_points = {
    }

    cmdclass = dict(
        build_doc=BuildDocCommand)

    config = configuration()
    setup_kwargs = config.todict()
    setup_kwargs.update(name=PROJECT,
                        version=VERSION,
                        url="https://gitlab.esrf.fr/denolf/oarpy",
                        author="Wout De Nolf",
                        author_email="wout.de_nolf@esrf.fr",
                        classifiers=classifiers,
                        description="Python binding for OAR",
                        long_description=get_readme(),
                        packages=PACKAGES,
                        install_requires=install_requires,
                        setup_requires=setup_requires,
                        cmdclass=cmdclass,
                        package_data=package_data,
                        zip_safe=False,
                        entry_points=entry_points,
                        )
    return setup_kwargs


def setup_package():
    """Run setup(**kwargs)

    Depending on the command, it either runs the complete setup which depends on numpy,
    or a *dry run* setup with no dependency on numpy.
    """

    # Check if action requires build/install
    dry_run = len(sys.argv) == 1 or (len(sys.argv) >= 2 and (
        '--help' in sys.argv[1:] or
        sys.argv[1] in ('--help-commands', 'egg_info', '--version',
                        'clean', '--name')))

    if dry_run:
        # DRY_RUN implies actions which do not require dependancies, like NumPy
        try:
            from setuptools import setup
            logger.info("Use setuptools.setup")
        except ImportError:
            from distutils.core import setup
            logger.info("Use distutils.core.setup")
    else:
        try:
            from setuptools import setup
        except ImportError:
            from numpy.distutils.core import setup
            logger.info("Use numpydistutils.setup")

    setup_kwargs = get_project_configuration(dry_run)
    setup(**setup_kwargs)


DATA_FILES = [
    # Data files that will be installed outside site-packages folder
]

if __name__ == "__main__":
    setup_package()
